#!/usr/bin/python
# -*- coding: utf-8 -*-

import argparse
import os
import sys
import re
import json
import random
import decimal
import fractions
import math
from collections import defaultdict

F_ONE = fractions.Fraction(1, 1)

class CollisionVariant(object):
    DEAD = 0
    BONET = 1
    BARTO = 2

    @staticmethod
    def from_string(c):
        cl = c.lower()
        return CollisionVariant.DEAD if cl == "dead" else \
                CollisionVariant.BONET if cl == "bonet" else \
                CollisionVariant.BARTO if cl == "barto" else None

class NoiseVariant(object):
    NOCHANGE = 0
    OFFBYONE = 1
    OFFBYONETRUNCATE = 2

    @staticmethod
    def from_string(c):
        cl = c.lower()
        return NoiseVariant.NOCHANGE if c == "nochange" else \
                NoiseVariant.OFFBYONE if c == "offbyone" else \
                NoiseVariant.OFFBYONETRUNCATE if c == "offbyonetrunc" else \
                None

class CellType(object):
    START = 0
    GOAL = 1
    WALL = 2
    EMPTY = 3
    UNDEFINED = 4

    @staticmethod
    def from_string(c):
        if c == 's':
            return CellType.START
        elif c == 'g':
            return CellType.GOAL
        elif c == 'x':
            return CellType.WALL
        elif c == '.':
            return CellType.EMPTY
        elif c == ' ':
            return CellType.UNDEFINED
        else:
            raise ValueError("Could parse cell type for string %r" % c)


class Grid(object):
    def __init__(self, dimx, dimy):
        self.grid = [ [CellType.EMPTY for y in range(dimy)] for x in range(dimx) ]
        self.start = []
        self.goal = []
        self.dimx = dimx
        self.dimy = dimy

    def set(self, x, y, t):
        self.grid[x][y] = t
        if t == CellType.START:
            self.start.append((x, y))
        elif t == CellType.GOAL:
            self.goal.append((x, y))

    def get(self, x, y):
        return self.grid[x][y]


def parse_grid_file(p, df, speed_limit):
    grid = None
    speed_limits = None
    with open(p) as f:
        lines = f.readlines()
        if len(lines) == 0:
            return None
        m = re.match("dim: (\d+) (\d+)", lines[0])
        assert(m != None)
        rows = int(m.group(1))
        cols = int(m.group(2))
        grid = Grid(df * int(cols), df * int(rows))
        speed_limits = [[0 for y in range(grid.dimy)] for x in range(grid.dimx)]
        lines = lines[1:]
        while len(lines) > int(m.group(1)):
            trim_last_new_lint = True
            assert(lines[-1].strip() == "")
            lines.pop()
        assert(len(lines) == rows)
        for y in range(rows):
            assert(len(lines[y]) >= cols)
            for x in range(cols):
                try:
                    t = CellType.EMPTY
                    speed = int(lines[y][x])
                except ValueError:
                    t = CellType.from_string(lines[y][x])
                    speed = 0
                for yd in range(y * df, y * df + df):
                    for xd in range(x * df, x * df + df):
                        grid.set(xd, yd, t)
                        speed_limits[xd][yd] = df * (speed if speed_limit is None else speed_limit)
    return (grid, speed_limits)


class Je(object):
    @staticmethod
    def _binary(var, val, op="="):
        return {"op": op, "left": var, "right": val}
    @staticmethod
    def Neg(exp):
        return Je.Sub(0, exp)
    @staticmethod
    def Abs(exp):
        return {"op": "abs", "exp": exp}
    @staticmethod
    def Min(x, y):
        return Je._binary(x, y, "min")
    @staticmethod
    def Max(x, y):
        return Je._binary(x, y, "max")
    @staticmethod
    def Eq(var, val):
        return Je._binary(var, val, "=")
    @staticmethod
    def Ne(var, val):
        return Je._binary(var, val, "≠")
    @staticmethod
    def Lt(var, val):
        return Je._binary(var, val, "<")
    @staticmethod
    def Le(var, val):
        return Je._binary(var, val, "≤")
    @staticmethod
    def Gt(var, val):
        return Je.Lt(val, var)
    @staticmethod
    def Ge(var, val):
        return Je.Le(val, var)
    @staticmethod
    def Add(var, val):
        return Je._binary(var, val, "+")
    @staticmethod
    def Sub(var, val):
        return Je._binary(var, val, "-")
    @staticmethod
    def Mult(var, val):
        return Je._binary(var, val, "*")
    @staticmethod
    def Div(var, val=None):
        if val is None:
            assert(isinstance(var, fractions.Fraction))
            return Je._binary(var.numerator, var.denominator)
        return Je._binary(var, val, "/")
    @staticmethod
    def Pow(x, y):
        return Je._binary(x, y, "pow")
    @staticmethod
    def concat(exps, i = 0, op="∧"):
        if i + 1 == len(exps):
            return exps[i]
        return {"op": op, "left": exps[i], "right": Je.concat(exps, i+1, op)}
    @staticmethod
    def And(*args):
        if len(args) == 1 and isinstance(args[0], list):
            return Je.concat(args[0], 0, "∧")
        return Je.concat(args, 0, "∧")
    @staticmethod
    def Or(*args):
        if len(args) == 1 and isinstance(args[0], list):
            return Je.concat(args[0], 0, "∨")
        return Je.concat(args, 0, "∨")
    @staticmethod
    def Not(exp):
        return {"op": "¬", "exp": exp}
    @staticmethod
    def Implies(cond, conseq):
        return Je.Or([Je.Not(cond), conseq])
    @staticmethod
    def Round(exp):
        return { "op": "floor", "exp": Je.Add(exp, 0.5) }
    # @staticmethod
    # def Round(exp):
    #     return Je.Ite(Je.Lt(exp, 0), Je.Sub(0, Je._Round(Je.Sub(0, exp))), Je._Round(exp))
    @staticmethod
    def MultRounded(var, val):
        return Je.Round(Je.Mult(var, val))
    @staticmethod
    def SubRounded(var, val):
        return Je.Round(Je.Sub(var, val))
    @staticmethod
    def AddRounded(var, val):
        return Je.Round(Je.Add(var, val))
    @staticmethod
    def ArrayConstruct(var, length, expr):
        return { "op": "ac", "var": var, "length": length, "exp": expr }
    @staticmethod
    def ArrayAccess(exp, ind):
        return { "op": "aa", "exp": exp, "index": ind }
    @staticmethod
    def ArrayValue(elem):
        return { "op": "av", "elements": elem }
    @staticmethod
    def Ite(i, t, e):
        return { "op": "ite", "if": i, "then": t, "else": e }
    @staticmethod
    def Value(var, val):
        return { "ref": var, "value": val }


class RacetrackGenerator(object):
    def __init__(self, one_dimensional, variant, parking, zero_velocity, noise_variant, initial_dx, initial_dy, noise_probability, storm_compatible):
        self.one_dimensional = one_dimensional
        self.collision_variant = variant
        self.parking = parking
        self.zero_velocity = zero_velocity
        self.noise_variant = noise_variant
        self.initial_dx = initial_dx
        self.initial_dy = initial_dy
        self.noise_probability = noise_probability
        self.storm_compatible = storm_compatible


    def _bounds(self):
        bounds = set()
        for x in range(0,self._grid.dimx):
            for y in range(0,self._grid.dimy):
                if self._grid.get(x,y) == CellType.EMPTY or self._grid.get(x,y) == CellType.START:
                    if x!=0:
                        if self._grid.get(x-1,y) == CellType.WALL:
                            bounds.add((x-1,y))
                    else:
                        bounds.add((x-1,y))
                    if x!=self._grid.dimx-1:
                        if self._grid.get(x+1,y) == CellType.WALL:
                            bounds.add((x+1,y))
                    else:
                            bounds.add((x+1,y))
                    if y!=0:
                        if self._grid.get(x,y-1) == CellType.WALL:
                            bounds.add((x,y-1))
                    else:
                        bounds.add((x,y-1))
                    if y!=self._grid.dimy-1:
                        if self._grid.get(x,y+1) == CellType.WALL:
                            bounds.add((x,y+1))
                    else:
                        bounds.add((x,y+1))
                    if x!=0 and y!=0:
                        if self._grid.get(x-1,y-1) == CellType.WALL:
                            bounds.add((x-1,y-1))
                    else:
                        bounds.add((x-1,y-1))
                    if x!=0 and  y!=self._grid.dimy-1:
                        if self._grid.get(x-1,y+1) == CellType.WALL:
                            bounds.add((x-1,y+1))
                    else:
                        bounds.add((x-1,y+1))   
                    if x!=self._grid.dimx-1 and y!=0:
                        if self._grid.get(x+1,y-1) == CellType.WALL: 
                            bounds.add((x+1,y-1))
                    else:
                        bounds.add((x+1,y-1))
                    if  x!=self._grid.dimx-1 and  y!=self._grid.dimy-1:
                        if self._grid.get(x+1,y+1) == CellType.WALL:
                            bounds.add((x+1,y+1))
                    else:
                        bounds.add((x+1,y+1))
        return bounds

    def _accel_name(self, car, ax, ay):
        sax = str(ax)
        say = str(ay)
        #if ax < 0:
        #    sax = "_%d" % (ax * -1)
        #if ay < 0:
        #    say = "_%d" % (ay * -1)
        return "accel%d_%s_%s" % (car, sax, say)

    def _actions(self):
        return [ { "name": self._accel_name(i, dx, dy) } for i in range(self._num_cars) for dy in [-1, 0, 1] for dx in [-1, 0, 1] ] \
            + [ { "name": "check_if_valid" } ] \
            + [ { "name": "is_valid" } ] \
            + [ { "name": "passed_finish_line" } ] \
            + [ { "name": "is_invalid" } ] \
            + [ { "name": "exceeded_speed_limit" } ] \
            + [ { "name": "synchronize" } ] \
            + [ { "name": "increment_step" } ] \
            + [ { "name": "reset" } ] \
            + [ { "name": "tau" } ]

    def _constants(self):
        return [
            { "name": "dim_x", "type": "int", "value": self._grid.dimx },
            { "name": "dim_y", "type": "int", "value": self._grid.dimy },
        ]

    def _variables(self, bounds):
        res = ([] if self._horizon is None else [ { "name": "horizon", "type": { "kind": "bounded", "base": "int", "lower-bound": 0, "upper-bound": self._horizon } } ])
        res += [
            # { "name": "all_finished", "type": "bool" },
            { "name": "step_indicator", "type": { "kind": "bounded", "base": "int", "lower-bound": 0, "upper-bound": 1 }, "initial-value": 0 },
            { "name": "step_completed", "type": "bool", "initial-value": True },

            { "name": "map_x", "type":  {"kind": "bounded", "base": "int", "lower-bound": -1, "upper-bound": self._grid.dimx } },
            { "name": "map_y", "type":  {"kind": "bounded", "base": "int", "lower-bound": -1, "upper-bound": self._grid.dimy } },
            { "name": "map_dx", "type": {"kind": "bounded", "base": "int", "lower-bound": -self._grid.dimx - 1, "upper-bound": self._grid.dimx + 1 } },
            { "name": "map_dy", "type": {"kind": "bounded", "base": "int", "lower-bound": -self._grid.dimy - 1, "upper-bound": self._grid.dimy + 1 } },

            { "name": "goal_x", "type":  {"kind": "bounded", "base": "int", "lower-bound": -1, "upper-bound": self._grid.dimx } },
            { "name": "goal_y", "type":  {"kind": "bounded", "base": "int", "lower-bound": -1, "upper-bound": self._grid.dimy } },
            { "name": "crash_x", "type":  {"kind": "bounded", "base": "int", "lower-bound": -2, "upper-bound": self._grid.dimx+1 } },
            { "name": "crash_y", "type":  {"kind": "bounded", "base": "int", "lower-bound": -2, "upper-bound": self._grid.dimy+1 } },
        ]
        for i in range(self._num_cars):
            # res.append({ "name": "car%d_startpos_status" % i, "type": {"kind": "bounded", "base": "int", "lower-bound": 0, "upper-bound": 2 } })
            res.append({ "name": "car%d_startpos_x" % i, "type": {"kind": "bounded", "base": "int", "lower-bound": -1, "upper-bound": self._grid.dimx } })
            res.append({ "name": "car%d_startpos_y" % i, "type": {"kind": "bounded", "base": "int", "lower-bound": -1, "upper-bound": self._grid.dimy } })
            res.append({ "name": "car%d_dx" % i, "type": {"kind": "bounded", "base": "int", "lower-bound": -self._grid.dimx - 1, "upper-bound": self._grid.dimx + 1 } })
            res.append({ "name": "car%d_dy" % i, "type": {"kind": "bounded", "base": "int", "lower-bound": -self._grid.dimy - 1, "upper-bound": self._grid.dimy + 1 } })
            res.append({ "name": "car%d_x" % i, "type":  {"kind": "bounded", "base": "int", "lower-bound": 0, "upper-bound": self._grid.dimx - 1 } })
            res.append({ "name": "car%d_y" % i, "type":  {"kind": "bounded", "base": "int", "lower-bound": 0, "upper-bound": self._grid.dimy - 1 } })
            res.append({ "name": "car%d_finished" % i, "type": "bool" })
            # for x,y in self._grid.goal:
            #     res.append({ "name": "car%d_finished_in_%d_%d" % (i,x,y), "type": "bool" })
            if self.collision_variant == CollisionVariant.DEAD:
                res.append({"name": "car%d_crashed" % i, "type": "bool" })
                # for x,y in bounds:
                #     res.append({"name": "car%d_crashed_in_%d_%d" % (i,x,y), "type": "bool" })
            res.append({"name": "car%d_is_deadlocked" % i, "type": "bool" })
        return res

    def _initial_state(self, bounds):
        res = [
                # Je.Eq("all_finished", False),
                Je.Eq("step_indicator", 0),
                Je.Eq("step_completed", True),
                Je.Eq("map_x", 0),
                Je.Eq("map_y", 0),
                Je.Eq("map_dx", 0),
                Je.Eq("map_dy", 0),
            ]
        if self._horizon != None:
            res.append(Je.Eq("horizon", 0))
        for i in range(self._num_cars):
            # res.append(Je.Eq("car%d_startpos_status" % i, 0))
            res.append(Je.Eq("car%d_startpos_x" % i, -1))
            res.append(Je.Eq("car%d_startpos_y" % i, -1))
            res.append(Je.Eq("car%d_x" % i, self._grid.start[i % len(self._grid.start)][0]))
            res.append(Je.Eq("car%d_y" % i, self._grid.start[i % len(self._grid.start)][1]))
            res.append(Je.Eq("car%d_dx" % i, self.initial_dx))
            res.append(Je.Eq("car%d_dy" % i, self.initial_dy))
            res.append(Je.Eq("car%d_finished" % i, False))
            # for x,y in self._grid.goal:
            #     res.append(Je.Eq("car%d_finished_in_%d_%d" % (i,x,y), False))
            if self.collision_variant == CollisionVariant.DEAD:
                 res.append(Je.Eq("car%d_crashed" % i, False))
            res.append(Je.Eq("car%d_is_deadlocked" % i, False))
            # if self.collision_variant == CollisionVariant.DEAD:
            #     for x,y in bounds:
            #         res.append(Je.Eq("car%d_crashed_in_%d_%d" % (i,x,y), False))
            res.append(Je.Eq("goal_x", -1))
            res.append(Je.Eq("goal_y", -1))
            res.append(Je.Eq("crash_x", -2))
            res.append(Je.Eq("crash_y", -2))
        return res

    def _all_finished(self):
        #if self.zero_velocity:
        #    return Je.And(["all_finished"] + [Je.Eq("car%d_dx" % i, 0) for i in range(self._num_cars)] + [Je.Eq("car%d_dy" % i, 0) for i in range(self._num_cars)])
        # return "all_finished"
        if self.zero_velocity:
            return Je.And(
                    ["car%d_finished" for i in range(1)]
                    + [Je.Eq("car%d_d%s" % (i, d), 0) for i in range(1) for d in ["x", "y"]])
        else:
            return "car0_finished"

    def _property_goal_probability(self):
        return [{
                "name": "goal_prob",
                "expression": {
                    "op": "filter",
                    "fun": "min",
                    "values": {
                        "op": "Pmax",
                        "exp": {
                            "op": "U",
                            "left": True,
                            "right": self._all_finished()
                            # "right": [Je.And(Je.Eq("car0_startpos_x", sx), Je.Eq("car0_startpos_y", sy), self._all_finished()) for (sx, sy) in self._grid.start][0],
                        }
                    },
                    # "states": [Je.And(Je.Eq("car0_startpos_status", 1), Je.Eq("car0_x", sx), Je.Eq("car0_y", sy)) for (sx, sy) in self._grid.start][0]
                    "states": { "op": "initial" }
                }
            }]


    def _property_goal_probability_in(self, x, y):
        return [{
                "name": "goal_prob_in_%d_%d" % (x,y),
                "expression": {
                    "op": "filter",
                    "fun": "min",
                    "values": {
                        "op": "Pmax",
                        "exp": {
                            "op": "U",
                            "left": Je.And(Je.Not("car0_finished"), Je.Not("car0_crashed")),
                            "right": Je.And(Je.Eq("goal_x", x), Je.Eq("goal_y", y))
                            # "right": [Je.And(Je.Eq("car0_startpos_x", sx), Je.Eq("car0_startpos_y", sy), self._all_finished()) for (sx, sy) in self._grid.start][0],
                        }
                    },
                    # "states": [Je.And(Je.Eq("car0_startpos_status", 1), Je.Eq("car0_x", sx), Je.Eq("car0_y", sy)) for (sx, sy) in self._grid.start][0]
                    "states": { "op": "initial" }
                }
            }]

    def _property_speed_exceeded(self):
        return [{
                "name": "speed_exceeded_%s_prob" % x,
                "expression": {
                    "op": "filter",
                    "fun": "min",
                    "values": {
                        "op": "P%s" % x,
                        "exp": {
                            "op": "U",
                            "left": Je.Not("car0_finished"),
                            "right": "car0_is_deadlocked"
                        }
                    },
                    "states": { "op": "initial" }
                }
            } for x in ["min", "max"]]

    def _property_deadlocked(self):
        if self.collision_variant != CollisionVariant.DEAD:
            return self._property_speed_exceeded()
        return [{
                "name": "deadlocked_%s_prob" % x,
                "expression": {
                    "op": "filter",
                    "fun": "min",
                    "values": {
                        "op": "P%s" % x,
                        "exp": {
                            "op": "U",
                            "left": Je.Not("car0_finished"),
                            "right": Je.Or("car0_is_deadlocked", "car0_crashed")
                        }
                    },
                    "states": { "op": "initial" }
                }
            } for x in ["min", "max"]]


    def _property_crashed_in(self, x_pos, y_pos):
        if self.collision_variant != CollisionVariant.DEAD:
            return []
        return [{
                "name": "crashed_%s_prob_in_%d_%d" % ("min",x_pos,y_pos),   #x statt "min"
                "expression": {
                    "op": "filter",
                    "fun": "min",
                    "values": {
                        "op": "Pmin", #"P%s" % x,
                        "exp": {
                            "op": "U",
                            "left": Je.And(Je.Not("car0_crashed"),Je.Not("car0_finished")),
                            "right": Je.And(Je.Eq("crash_x", x_pos), Je.Eq("crash_y", y_pos))
                        }
                    },
                    "states": { "op": "initial" }
                }
            } #for x in ["min", "max"]
        ]

    def _property_crashed(self):
        if self.collision_variant != CollisionVariant.DEAD:
            return []
        return [{
                "name": "crashed_%s_prob" % x,
                "expression": {
                    "op": "filter",
                    "fun": "min",
                    "values": {
                        "op": "P%s" % x,
                        "exp": {
                            "op": "U",
                            "left": Je.Not("car0_finished"),
                            "right": "car0_crashed"
                        }
                    },
                    "states": { "op": "initial" }
                }
            } for x in ["min", "max"]]

    def _property_no_dead_ends(self):
        return [{
                "name": "no_dead_ends",
                "expression": {
                    "op": "filter",
                    "fun": "∀",
                    "values": {
                        "op": "∀",
                        "exp": {
                            "op": "U",
                            "left": True,
                            "right": self._all_finished()
                        }
                    },
                    "states": {"op": "initial" }
                }
            }]

    def _property_expected_steps(self):
        return [{
                "name": "expected_steps",
                "expression": {
                    "op": "filter",
                    "fun": "max",
                    "values": {
                        "op": "Emin",
                        "exp": "step_indicator",
                        "accumulate": [ "exit" ],
                        "reach": Je.And("step_completed", self._all_finished())
                    },
                    "states": { "op": "initial" }
                }
            }]

    def _properties(self, bounds):
        # return [self._property_goal_probability()] + ([] if self.collision_variant == CollisionVariant.DEAD else [self._property_expected_steps()])
        crashProbs = []
        for x,y in bounds:
            crashProbs.extend(self._property_crashed_in(x,y))
        goalProbs = []
        for x,y in self._grid.goal:
            goalProbs.extend(self._property_goal_probability_in(x,y))
        return crashProbs + goalProbs + self._property_goal_probability() # + [self._property_no_dead_ends()]
        #return self._property_goal_probability() + self._property_expected_steps() + self._property_crashed() + self._property_speed_exceeded() + self._property_deadlocked() + crashProbs + goalProbs # + [self._property_no_dead_ends()]

    def _map_automaton_discrete(self):
        grid_var = { "name": "grid", "type": { "kind": "array", "base": { "kind": "array", "base": { "kind": "bounded", "base": "int", "lower-bound": 0, "upper-bound": 2 } } } }
        grid_var_value = Je.Eq("grid", Je.ArrayValue([Je.ArrayValue([1 if self._grid.get(x, y) == CellType.WALL else (2 if self._grid.get(x,y) == CellType.GOAL else 0) for y in range(self._grid.dimy)]) for x in range(self._grid.dimx)]))
        speed_limits_var = { "name": "speed_limits", "type": { "kind": "array", "base": { "kind": "array", "base": { "kind": "bounded", "base": "int", "lower-bound": 0, "upper-bound": 9 } } } }
        speed_limits_var_value = Je.Eq("speed_limits", Je.ArrayValue([Je.ArrayValue([self._speed_limits[x][y] for y in range(self._grid.dimy)]) for x in range(self._grid.dimx)]))
        if self.one_dimensional:
            grid_var = { "name": "grid",
                    "type": { "kind": "array", "base":
                        { "kind": "bounded", "base": "int", "lower-bound": 0, "upper-bound": 2 } },
                    }
            speed_limits_var = { "name": "speed_limits",
                    "type": { "kind": "array", "base":
                        { "kind": "bounded", "base": "int", "lower-bound": 0, "upper-bound": 9 } },
                    }
            if self.storm_compatible:
                grid_var["initial-value"] = Je.ArrayValue([1 if self._grid.get(x, y) == CellType.WALL else (2 if self._grid.get(x,y) == CellType.GOAL else 0) for y in range(self._grid.dimy) for x in range(self._grid.dimx)])
                speed_limits_var["initial-value"] = Je.ArrayValue([self._speed_limits[x][y] for y in range(self._grid.dimy) for x in range(self._grid.dimx)])
            grid_var_value = Je.Eq("grid", Je.ArrayValue([1 if self._grid.get(x, y) == CellType.WALL else (2 if self._grid.get(x,y) == CellType.GOAL else 0) for y in range(self._grid.dimy) for x in range(self._grid.dimx)]))
            speed_limits_var_value = Je.Eq("speed_limits", Je.ArrayValue([self._speed_limits[x][y] for y in range(self._grid.dimy) for x in range(self._grid.dimx)]))

        map_aut = {
            "name": "map",
            "locations": [
                { "name": "map_dx_check_intersection",  },
                { "name": "map_dx_check_speed_limit",  },
                { "name": "map_dx_check_x_bounds",  },
                { "name": "map_dxy_stepx_check_intersection_update_y",  },
                { "name": "map_dxy_stepx_check_speed_limit",  },
                { "name": "map_dxy_stepx_check_x_bounds",  },
                { "name": "map_dxy_stepx_check_y_bounds",  },
                { "name": "map_dxy_stepy_check_intersection_update_x",  },
                { "name": "map_dxy_stepy_check_speed_limit",  },
                { "name": "map_dxy_stepy_check_x_bounds",  },
                { "name": "map_dxy_stepy_check_y_bounds",  },
                { "name": "map_dy_check_intersection",  },
                { "name": "map_dy_check_speed_limit",  },
                { "name": "map_dy_check_y_bounds",  },
                { "name": "map_exceeded_speed_limit",  },
                { "name": "map_initialize_steps",  },
                { "name": "map_invalid_trajectory",  },
                { "name": "map_trajectory_passed_goal",  },
                { "name": "map_wait_for_trajectory",  },
            ],
            "variables": [
                grid_var,
                speed_limits_var,
                { "name": "step_increment", "type": { "kind": "bounded", "base": "int", "lower-bound": -1, "upper-bound": 1 }, "initial-value": 0 },
                { "name": "next_pos", "type": "int" }, # { "kind": "bounded", "base": "int", "lower-bound": -self._grid.dimy, "upper-bound": 2*self._grid.dimy }, "initial-value": 0 },
                { "name": "start_x", "type": { "kind": "bounded", "base": "int", "lower-bound": 0, "upper-bound": self._grid.dimx }, "initial-value": 0 },
                { "name": "start_y", "type": { "kind": "bounded", "base": "int", "lower-bound": 0, "upper-bound": self._grid.dimy }, "initial-value": 0 },
                { "name": "end_x", "type": { "kind": "bounded", "base": "int", "lower-bound": -2 * self._grid.dimx, "upper-bound": 3*self._grid.dimx }, "initial-value": 0 },
                { "name": "end_y", "type": { "kind": "bounded", "base": "int", "lower-bound": -2 * self._grid.dimy, "upper-bound": 3*self._grid.dimy }, "initial-value": 0 },
                { "name": "current_speed", "type": { "kind": "bounded", "base": "int", "lower-bound": 0, "upper-bound": (self._grid.dimx + self._grid.dimy)**2 }, "initial-value": 0 },
            ],
            "restrict-initial": {
                "exp": Je.And(([] if self.storm_compatible else [grid_var_value, speed_limits_var_value]) + [
                    Je.Eq("step_increment", 0),
                    Je.Eq("next_pos", 0),
                    Je.Eq("start_x", 0),
                    Je.Eq("end_x", 0),
                    Je.Eq("start_y", 0),
                    Je.Eq("end_y", 0),
                    Je.Eq("current_speed", 0),
                    ])
            },
            "initial-locations": [ "map_wait_for_trajectory" ],
            "edges" : [
                {
                    "location": "map_invalid_trajectory",
                    "action": "is_invalid",
                    "destinations": [ { "location": "map_wait_for_trajectory", } ]
                },
                {
                    "location": "map_exceeded_speed_limit",
                    "action": "exceeded_speed_limit",
                    "destinations": [ { "location": "map_wait_for_trajectory", } ]
                },
                {
                    "location": "map_trajectory_passed_goal",
                    "action": "passed_finish_line",
                    "destinations": [ { "location": "map_wait_for_trajectory", } ]
                },
            ]
        }

        def goal_intersection(xvar="map_x", yvar="map_y"):
            # return Je.And(Je.Lt(1, "map_x"), Je.Le(0, "map_y"), Je.Lt(yvar, "dim_y"))
            if self.one_dimensional:
                return Je.Eq(Je.ArrayAccess("grid", Je.Add(xvar, Je.Mult(self._grid.dimx, yvar))), 2)
            else:
                return Je.Eq(Je.ArrayAccess(Je.ArrayAccess("grid", xvar), yvar), 2)

        def wall_intersection(xvar="map_x", yvar="map_y"):
            # return Je.Not(Je.And(Je.Le(0, "map_x"), Je.Lt("map_x", "dim_x"), Je.Le(0, yvar), Je.Lt(yvar, "dim_y")))
            if self.one_dimensional:
                return Je.Eq(Je.ArrayAccess("grid", Je.Add(xvar, Je.Mult(self._grid.dimx, yvar))), 1)
            else:
                return Je.Eq(Je.ArrayAccess(Je.ArrayAccess("grid", xvar), yvar), 1)

        def intersection(xvar="map_x", yvar="map_y"):
            if self.parking:
                return wall_intersection(xvar, yvar)
            else:
                return Je.Or([wall_intersection(xvar, yvar), goal_intersection(xvar, yvar)])

        def speed_limit(xvar="map_x", yvar="map_y"):
            if self.one_dimensional:
                return Je.Pow(Je.ArrayAccess("speed_limits", Je.Add(xvar, Je.Mult(self._grid.dimx, yvar))), 2)
            else:
                return Je.Pow(Je.ArrayAccess(Je.ArrayAccess("speed_limits", xvar), yvar), 2)

        slope_x = Je.Div(Je.Sub("end_y", "start_y"), Je.Sub("end_x", "start_x"))
        offset_x = Je.Sub("end_y", Je.Mult("end_x", slope_x))

        slope_y = Je.Div(Je.Sub("end_x", "start_x"), Je.Sub("end_y", "start_y"))
        offset_y = Je.Sub("end_x", Je.Mult("end_y", slope_y))

        current_speed = "current_speed"

        reset_vars = [
            { "ref": "step_increment", "value": 0},
            { "ref": "start_x", "value": 0},
            { "ref": "end_x", "value": 0},
            { "ref": "start_y", "value": 0},
            { "ref": "end_y", "value": 0},
            { "ref": "next_pos", "value": 0},
            { "ref": "current_speed", "value": 0},
        ]

        map_aut["edges"].append({
            "location": "map_wait_for_trajectory",
            "action": "check_if_valid",
            "destinations": [ {
                "location": "map_initialize_steps",
                "assignments": [
                ] + reset_vars
            }]
        })

        map_aut["edges"].append({
            "location": "map_initialize_steps",
            "guard": { "exp": Je.And(Je.Eq("map_dx", 0), Je.Eq("map_dy", 0)) },
            "destinations": [ {
                "location": "map_dy_check_intersection",
                "assignments": [
                    { "ref": "step_increment", "value": 0 },
                ]
            }]
        })

        map_aut["edges"].append({
            "location": "map_initialize_steps",
            "guard": { "exp": Je.Or(
                Je.And(Je.Lt("map_dx", 0), Je.Lt("map_dy", 0), Je.Le("map_dx", "map_dy")),
                Je.And(Je.Lt("map_dx", 0), Je.Gt("map_dy", 0), Je.Ge(Je.Sub(0, "map_dx"), "map_dy")),
                Je.And(Je.Gt("map_dx", 0), Je.Lt("map_dy", 0), Je.Ge("map_dx", Je.Sub(0, "map_dy"))),
                Je.And(Je.Gt("map_dx", 0), Je.Gt("map_dy", 0), Je.Ge("map_dx", "map_dy"))) },
            "destinations": [ {
                "location": "map_dxy_stepx_check_intersection_update_y",
                "assignments": [
                    { "ref": "step_increment", "value": Je.Ite(Je.Lt("map_dx", 0), -1, 1) },
                    { "ref": "next_pos", "value": "map_y" },
                    { "ref": "map_dy", "value": 0},
                    { "ref": "start_x", "value": "map_x" },
                    { "ref": "start_y", "value": "map_y" },
                    { "ref": "end_x", "value": Je.Add("map_x", "map_dx") },
                    { "ref": "end_y", "value": Je.Add("map_y", "map_dy") },
                    { "ref": "current_speed", "value": Je.Add(Je.Pow("map_dx", 2), Je.Pow("map_dy", 2)) },
                ]
            }]
        })

        map_aut["edges"].append({
            "location": "map_initialize_steps",
            "guard": { "exp": Je.Or(
                Je.And(Je.Lt("map_dx", 0), Je.Lt("map_dy", 0), Je.Gt("map_dx", "map_dy")),
                Je.And(Je.Lt("map_dx", 0), Je.Gt("map_dy", 0), Je.Lt(Je.Sub(0, "map_dx"), "map_dy")),
                Je.And(Je.Gt("map_dx", 0), Je.Lt("map_dy", 0), Je.Lt("map_dx", Je.Sub(0, "map_dy"))),
                Je.And(Je.Gt("map_dx", 0), Je.Gt("map_dy", 0), Je.Lt("map_dx", "map_dy"))) },
            "destinations": [ {
                "location": "map_dxy_stepy_check_intersection_update_x",
                "assignments": [
                    { "ref": "step_increment", "value": Je.Ite(Je.Lt("map_dy", 0), -1, 1) },
                    { "ref": "next_pos", "value": "map_x" },
                    { "ref": "map_dx", "value": 0},
                    { "ref": "start_x", "value": "map_x" },
                    { "ref": "start_y", "value": "map_y" },
                    { "ref": "end_x", "value": Je.Add("map_x", "map_dx") },
                    { "ref": "end_y", "value": Je.Add("map_y", "map_dy") },
                    { "ref": "current_speed", "value": Je.Add(Je.Pow("map_dx", 2), Je.Pow("map_dy", 2)) },
                ]
            }]
        })

        map_aut["edges"].append({
            "location": "map_initialize_steps",
            "guard": { "exp": Je.And(Je.Eq("map_dx", 0), Je.Ne("map_dy", 0)) },
            "destinations": [ {
                "location": "map_dy_check_intersection",
                "assignments": [
                    { "ref": "step_increment", "value": Je.Ite(Je.Lt("map_dy", 0), -1, 1) },
                    { "ref": "current_speed", "value": Je.Add(Je.Pow("map_dx", 2), Je.Pow("map_dy", 2)) },
                ]
            }]
        })

        map_aut["edges"].append({
            "location": "map_initialize_steps",
            "guard": { "exp": Je.And(Je.Ne("map_dx", 0), Je.Eq("map_dy", 0)) },
            "destinations": [ {
                "location": "map_dx_check_intersection",
                "assignments": [
                    { "ref": "step_increment", "value": Je.Ite(Je.Lt("map_dx", 0), -1, 1) },
                    { "ref": "current_speed", "value": Je.Add(Je.Pow("map_dx", 2), Je.Pow("map_dy", 2)) },
                ]
            }]
        })

        def finished_all_steps_edge(loc, var):
            if self.parking:
                return [
                {
                    "location": loc,
                    "action": "is_valid",
                    "guard": { "exp": Je.And(Je.Not(goal_intersection()), Je.Eq(var, 0)) },
                    "destinations": [ {
                        "location": "map_wait_for_trajectory",
                        "assignments": reset_vars
                    }]
                },
                {
                    "location": loc,
                    "action": "passed_finish_line",
                    "guard": { "exp": Je.And(goal_intersection(), Je.Eq(var, 0)) },
                    "destinations": [ {
                        "location": "map_wait_for_trajectory",
                        "assignments": reset_vars
                    }]
                }]
            else:
                return [{
                    "location": loc,
                    "action": "is_valid",
                    "guard": { "exp": Je.Eq(var, 0) },
                    "destinations": [ {
                        "location": "map_wait_for_trajectory",
                        "assignments": reset_vars
                    }]
                }]

        ### dx!=0 && dy!=0
        map_aut["edges"] += finished_all_steps_edge("map_dxy_stepx_check_x_bounds", "map_dx")

        map_aut["edges"] += [{
            "location": "map_dxy_stepx_check_x_bounds",
            "guard": { "exp": Je.And(Je.Ne("map_dx", 0), Je.Or(Je.Lt(Je.Add("map_x", "step_increment"), 0), Je.Le("dim_x", Je.Add("map_x", "step_increment")))) },
            "destinations": [ {
                "location": "map_invalid_trajectory",
                "assignments":  [
                    { "ref": "crash_x", "value": Je.Add("map_x", "step_increment") },
                    { "ref": "crash_y", "value": "map_y" },
                ] + reset_vars
            }]
        },
        {
            "location": "map_dxy_stepx_check_x_bounds",
            "guard": { "exp": Je.And(Je.Ne("map_dx", 0), Je.Le(0, Je.Add("map_x", "step_increment")), Je.Lt(Je.Add("map_x", "step_increment"), "dim_x")) },
            "destinations": [ {
                "location": "map_dxy_stepx_check_y_bounds",
                "assignments": [
                    { "ref": "map_x", "value": Je.Add("map_x", "step_increment") },
                    { "ref": "map_dx", "value": Je.Sub("map_dx", "step_increment") },
                    { "ref": "next_pos", "value": Je.Round(Je.Add(Je.Mult(Je.Add("map_x", "step_increment"), slope_x), offset_x)) },
                ]
            }]
        }]

        map_aut["edges"] += [{
            "location": "map_dxy_stepx_check_y_bounds",
            "guard": { "exp": Je.Or(Je.Lt("next_pos", 0), Je.Le("dim_y", "next_pos")) },
            "destinations": [ {
                "location": "map_invalid_trajectory",
                "assignments": [
                    { "ref": "crash_x", "value": "map_x" },
                    { "ref": "crash_y", "value": "next_pos" },
                    { "ref": "map_x", "value": Je.Sub("map_x", "step_increment") },
                ] + reset_vars
            }]
        },
        {
            "location": "map_dxy_stepx_check_y_bounds",
            "guard": { "exp": Je.And(Je.Le(0, "next_pos"), Je.Lt("next_pos", "dim_y")) },
            "destinations": [ {
                "location": "map_dxy_stepx_check_intersection_update_y",
            }]
        }]

        map_aut["edges"] += [{
            "location": "map_dxy_stepx_check_intersection_update_y",
            "guard": { "exp": Je.Not(intersection(yvar="next_pos")) } ,
            "destinations": [ {
                # "location": "map_dxy_stepx_check_x_bounds",
                "location": "map_dxy_stepx_check_speed_limit",
                "assignments": [
                    { "ref": "map_y", "value": "next_pos" },
                    { "ref": "next_pos", "value": 0 }
                ]
            }]
        },
        {
            "location": "map_dxy_stepx_check_intersection_update_y",
            "guard": { "exp": wall_intersection(yvar="next_pos") } ,
            "destinations": [ {
                "location": "map_invalid_trajectory",
                "assignments": reset_vars + [
                    { "ref": "crash_x", "value": "map_x" },
                    { "ref": "crash_y", "value": "next_pos" },
                    { "ref": "map_x", "value": Je.Sub("map_x", "step_increment") }, # last valid position
                ]
            }]
        }]
        if not self.parking:
            map_aut["edges"].append({ # is a goal cell -> passed goal line
                "location": "map_dxy_stepx_check_intersection_update_y",
                "guard": { "exp": goal_intersection(yvar="next_pos") } ,
                "destinations": [ {
                    "location": "map_trajectory_passed_goal",
                    "assignments": reset_vars + [
                        { "ref": "goal_x", "value": "map_x" },
                        { "ref": "goal_y", "value": "next_pos" },
                        { "ref": "map_y", "value": "next_pos" }
                    ]
                }]
            })
        map_aut["edges"].append({
            "location": "map_dxy_stepx_check_speed_limit",
            "guard": { "exp": Je.Or(Je.Eq(speed_limit(), 0), Je.Le(current_speed, speed_limit())) },
            "destinations": [ {
                "location": "map_dxy_stepx_check_x_bounds",
            }]
        })
        map_aut["edges"].append({
            "location": "map_dxy_stepx_check_speed_limit",
            "guard": { "exp": Je.And(Je.Gt(speed_limit(), 0), Je.Gt(current_speed, speed_limit())) },
            "destinations": [ {
                "location": "map_exceeded_speed_limit",
                "assignments": reset_vars,
            }]
        })

        ### dx!=0 && dy!=0
        map_aut["edges"] += finished_all_steps_edge("map_dxy_stepy_check_y_bounds", "map_dy")

        map_aut["edges"] += [{
            "location": "map_dxy_stepy_check_y_bounds",
            "guard": { "exp": Je.And(Je.Ne("map_dy", 0), Je.Or(Je.Lt(Je.Add("map_y", "step_increment"), 0), Je.Le("dim_y", Je.Add("map_y", "step_increment")))) },
            "destinations": [ {
                "location": "map_invalid_trajectory",
                "assignments": [
                    { "ref": "crash_x", "value": "map_x" },
                    { "ref": "crash_y", "value":  Je.Add("map_y", "step_increment") },
                ] + reset_vars
            }]
        },
        {
            "location": "map_dxy_stepy_check_y_bounds",
            "guard": { "exp": Je.And(Je.Ne("map_dy", 0), Je.Le(0, Je.Add("map_y", "step_increment")), Je.Lt(Je.Add("map_y", "step_increment"), "dim_y")) },
            "destinations": [ {
                "location": "map_dxy_stepy_check_x_bounds",
                "assignments": [
                    { "ref": "map_y", "value": Je.Add("map_y", "step_increment") },
                    { "ref": "map_dy", "value": Je.Sub("map_dy", "step_increment") },
                    { "ref": "next_pos", "value": Je.Round(Je.Add(Je.Mult(Je.Add("map_y", "step_increment"), slope_y), offset_y)) },
                ]
            }]
        }]

        map_aut["edges"] += [{
            "location": "map_dxy_stepy_check_x_bounds",
            "guard": { "exp": Je.Or(Je.Lt("next_pos", 0), Je.Le("dim_x", "next_pos")) },
            "destinations": [ {
                "location": "map_invalid_trajectory",
                "assignments": [
                    { "ref": "map_y", "value": Je.Sub("map_y", "step_increment") },
                    { "ref": "crash_x", "value": "next_pos" },
                    { "ref": "crash_y", "value": "map_y" },
                ] + reset_vars
            }]
        },
        {
            "location": "map_dxy_stepy_check_x_bounds",
            "guard": { "exp": Je.And(Je.Le(0, "next_pos"), Je.Lt("next_pos", "dim_x")) },
            "destinations": [ {
                "location": "map_dxy_stepy_check_intersection_update_x",
            }]
        }]

        map_aut["edges"] += [{
            "location": "map_dxy_stepy_check_intersection_update_x",
            "guard": { "exp": Je.Not(intersection(xvar="next_pos")) } ,
            "destinations": [ {
                "location": "map_dxy_stepy_check_speed_limit",
                "assignments": [
                    { "ref": "map_x", "value": "next_pos" },
                    { "ref": "next_pos", "value": 0 }
                ]
            }]
        },
        {
            "location": "map_dxy_stepy_check_intersection_update_x",
            "guard": { "exp": wall_intersection(xvar="next_pos") } ,
            "destinations": [ {
                "location": "map_invalid_trajectory",
                "assignments": reset_vars + [
                    { "ref": "crash_x", "value": "next_pos" },
                    { "ref": "crash_y", "value": "map_y" },
                    { "ref": "map_y", "value": Je.Sub("map_y", "step_increment") }, # last valid position
                ]
            }]
        }]
        if not self.parking:
            map_aut["edges"].append({ # is a goal cell -> passed goal line
                "location": "map_dxy_stepy_check_intersection_update_x",
                "guard": { "exp": goal_intersection(xvar="next_pos") } ,
                "destinations": [ {
                    "location": "map_trajectory_passed_goal",
                    "assignments": reset_vars + [
                        { "ref": "goal_x", "value": "next_pos" },
                        { "ref": "goal_y", "value": "map_y" },
                        { "ref": "map_x", "value": "next_pos" }
                    ]
                }]
            })
        map_aut["edges"].append({
            "location": "map_dxy_stepy_check_speed_limit",
            "guard": { "exp": Je.Or(Je.Eq(speed_limit(), 0), Je.Le(current_speed, speed_limit())) },
            "destinations": [ {
                "location": "map_dxy_stepy_check_y_bounds",
            }]
        })
        map_aut["edges"].append({
            "location": "map_dxy_stepy_check_speed_limit",
            "guard": { "exp": Je.And(Je.Gt(speed_limit(), 0), Je.Gt(current_speed, speed_limit())) },
            "destinations": [ {
                "location": "map_exceeded_speed_limit",
                "assignments": reset_vars,
            }]
        })



        ### dx==0 && dy!=0
        map_aut["edges"] += finished_all_steps_edge("map_dy_check_y_bounds", "map_dy")

        map_aut["edges"] += [
            { # out of bounds:
                "location": "map_dy_check_y_bounds",
                "guard": { "exp": Je.And(Je.Ne("map_dy", 0), Je.Or(Je.Lt(Je.Add("map_y", "step_increment"), 0), Je.Le("dim_y", Je.Add("map_y", "step_increment")))) },
                "destinations": [ {
                    "location": "map_invalid_trajectory",
                    "assignments": [
                        { "ref": "crash_x", "value": "map_x" },
                        { "ref": "crash_y", "value": Je.Add("map_y", "step_increment") },
                    ] + reset_vars
                }]
            },
            { # within bounds, update coordinates and go to "grid lookup" location
                "location": "map_dy_check_y_bounds",
                "guard": { "exp": Je.And(Je.Ne("map_dy", 0), Je.Le(0, Je.Add("map_y", "step_increment")), Je.Lt(Je.Add("map_y", "step_increment"), "dim_y")) },
                "destinations": [ {
                    "location": "map_dy_check_intersection",
                    "assignments": [
                        { "ref": "map_y", "value": Je.Add("map_y", "step_increment") },
                        { "ref": "map_dy", "value": Je.Sub("map_dy", "step_increment") },
                    ]
                }]
            }]

        map_aut["edges"] += [
        { # empty cell -> go back to loop location
            "location": "map_dy_check_intersection",
            "guard": { "exp": Je.Not(intersection()) } ,
            "destinations": [ {
                "location": "map_dy_check_speed_limit",
            }]
        },
        { # is a wall -> invalid move
            "location": "map_dy_check_intersection",
            "guard": { "exp": wall_intersection() } ,
            "destinations": [ {
                "location": "map_invalid_trajectory",
                "assignments": reset_vars + [
                    { "ref": "crash_x", "value": "map_x" },
                    { "ref": "crash_y", "value": "map_y" },
                    { "ref": "map_y", "value": Je.Sub("map_y", "step_increment") }, # last valid position
                ]
            }]
        }]
        if not self.parking:
            map_aut["edges"].append({ # is a goal cell -> passed goal line
                "location": "map_dy_check_intersection",
                "guard": { "exp": goal_intersection() } ,
                "destinations": [ {
                    "location": "map_trajectory_passed_goal",
                    "assignments": [
                        { "ref": "goal_x", "value": "map_x" },
                        { "ref": "goal_y", "value": "map_y" },
                    ] + reset_vars
                }]
            })
        map_aut["edges"].append({
            "location": "map_dy_check_speed_limit",
            "guard": { "exp": Je.Or(Je.Eq(speed_limit(), 0), Je.Le(current_speed, speed_limit())) },
            "destinations": [ {
                "location": "map_dy_check_y_bounds",
            }]
        })
        map_aut["edges"].append({
            "location": "map_dy_check_speed_limit",
            "guard": { "exp": Je.And(Je.Gt(speed_limit(), 0), Je.Gt(current_speed, speed_limit())) },
            "destinations": [ {
                "location": "map_exceeded_speed_limit",
                "assignments": reset_vars,
            }]
        })

        ### dx!=0 && dy==0
        map_aut["edges"] += finished_all_steps_edge("map_dx_check_x_bounds", "map_dx")

        map_aut["edges"] += [
        { # out of bounds:
            "location": "map_dx_check_x_bounds",
            "guard": { "exp": Je.And(Je.Ne("map_dx", 0), Je.Or(Je.Lt(Je.Add("map_x", "step_increment"), 0), Je.Le("dim_x", Je.Add("map_x", "step_increment")))) },
            "destinations": [ {
                "location": "map_invalid_trajectory",
                "assignments": [
                    { "ref": "crash_x", "value": Je.Add("map_x", "step_increment") },
                    { "ref": "crash_y", "value": "map_y" },
                ] + reset_vars
            }]
        },
        { # within bounds, update coordinates and go to "grid lookup" location
            "location": "map_dx_check_x_bounds",
            "guard": { "exp": Je.And(Je.Ne("map_dx", 0), Je.Le(0, Je.Add("map_x", "step_increment")), Je.Lt(Je.Add("map_x", "step_increment"), "dim_x")) },
            "destinations": [ {
                "location": "map_dx_check_intersection",
                "assignments": [
                    { "ref": "map_x", "value": Je.Add("map_x", "step_increment") },
                    { "ref": "map_dx", "value": Je.Sub("map_dx", "step_increment") },
                ]
            }]
        }]

        map_aut["edges"] += [
        { # empty cell -> go back to loop location
            "location": "map_dx_check_intersection",
            "guard": { "exp": Je.Not(intersection()) } ,
            "destinations": [ {
                "location": "map_dx_check_speed_limit",
            }]
        },
        { # is a wall -> invalid move
            "location": "map_dx_check_intersection",
            "guard": { "exp": wall_intersection() } ,
            "destinations": [ {
                "location": "map_invalid_trajectory",
                "assignments": reset_vars + [
                    { "ref": "crash_x", "value": "map_x" },
                    { "ref": "crash_y", "value": "map_y" },
                    { "ref": "map_x", "value": Je.Sub("map_x", "step_increment") }, # last valid position
                ]
            }]
        }]
        if not self.parking:
            map_aut["edges"].append({ # is a goal cell -> passed goal line
                "location": "map_dx_check_intersection",
                "guard": { "exp": goal_intersection() } ,
                "destinations": [ {
                    "location": "map_trajectory_passed_goal",
                    "assignments": [
                        { "ref": "goal_x", "value": "map_x" },
                        { "ref": "goal_y", "value": "map_y" },
                    ] + reset_vars
                }]
            })
        map_aut["edges"].append({
            "location": "map_dx_check_speed_limit",
            "guard": { "exp": Je.Or(Je.Eq(speed_limit(), 0), Je.Le(current_speed, speed_limit())) },
            "destinations": [ {
                "location": "map_dx_check_x_bounds",
            }]
        })
        map_aut["edges"].append({
            "location": "map_dx_check_speed_limit",
            "guard": { "exp": Je.And(Je.Gt(speed_limit(), 0), Je.Gt(current_speed, speed_limit())) },
            "destinations": [ {
                "location": "map_exceeded_speed_limit",
                "assignments": reset_vars,
            }]
        })

        return map_aut

    def _car_automaton(self, car_id):
        locstr = "car%d_%%s" % car_id

        car0x = "car%d_x" % car_id
        car0y = "car%d_y" % car_id
        car0dx = "car%d_dx" % car_id
        car0dy = "car%d_dy" % car_id
        car0finished = "car%d_finished" % car_id
        car0crashed = "car%d_crashed" % car_id
        car0deadlocked = "car%d_is_deadlocked" % car_id
        car0startposx = "car%d_startpos_x" % car_id
        car0startposy = "car%d_startpos_y" % car_id

        car0_aut = {
            "name": "car%d" % car_id,
            "locations": [ { "name": locstr % "choose_startpos" }, { "name": locstr % "wait_for_control" }, { "name": locstr % "send_trajectory_to_map" }, { "name": locstr % "wait_for_result" }, { "name": locstr % "sync_step" }, { "name": locstr % "deadlocked" } ],
            "initial-locations": [ locstr % "choose_startpos" if len(self._grid.start) > 1 else locstr % "wait_for_control" ],
            "edges": []
        }

        for (sx, sy) in self._grid.start:
            car0_aut["edges"] += [ {
                "location": locstr % "choose_startpos",
                "destinations": [ {
                    "location": locstr % "wait_for_control",
                    "assignments": [
                        { "ref": car0x, "value": sx },
                        { "ref": car0y, "value": sy },
                        { "ref": car0startposx, "value": sx },
                        { "ref": car0startposy, "value": sy },
                    ]
                }]
            } ]

        def _dest(p, ax, ay):
            l = locstr % "send_trajectory_to_map"
            res = { "location": l, "probability": { "exp": Je.Div(p.numerator, p.denominator) }, "assignments": [] }
            if ax != 0:
                res["assignments"].append({"ref": car0dx, "value": {"op": "+", "left": car0dx, "right": ax } })
            if ay != 0:
                res["assignments"].append({"ref": car0dy, "value": {"op": "+", "left": car0dy, "right": ay } })
            return res
        def _p(inv=False, x=1):
            if inv:
                p = (F_ONE - self.noise_probability) * fractions.Fraction(1, x)
            else:
                p = self.noise_probability * fractions.Fraction(1, x)
            return p
        if (self.noise_variant == NoiseVariant.OFFBYONE or self.noise_variant == NoiseVariant.OFFBYONETRUNCATE) \
                and self.noise_probability < F_ONE:
            for ax in [-1, 0, 1]:
                for ay in [-1, 0, 1]:
                    dests = defaultdict(lambda: fractions.Fraction(0))
                    dests[(ax, ay)] = _p()
                    noise_effects = [(dax, day) for dax in [-1, 0, 1] for day in [-1, 0, 1] if (dax != 0 or day != 0)]
                    for (dax, day) in noise_effects:
                        ddax = ax + dax
                        dday = ay + day
                        if self.noise_variant == NoiseVariant.OFFBYONETRUNCATE:
                            ddax = min(1, max(-1, ddax))
                            dday = min(1, max(-1, dday))
                        dests[(ddax, dday)] += _p(True, len(noise_effects))
                    car0_aut["edges"].append({
                        "location": locstr % "wait_for_control",
                        "action": self._accel_name(car_id, ax, ay),
                        "guard": {
                            "exp": (Je.Not(car0crashed) if self.collision_variant == CollisionVariant.DEAD else True)
                        },
                        "destinations": [_dest(dests[(dax, day)], dax, day) for (dax, day) in dests]})
        else:
            car0_aut["edges"] += [ {
                "location": locstr % "wait_for_control",
                "action": self._accel_name(car_id, ax, ay),
                "guard": {
                    "exp": (Je.Not(car0crashed) if self.collision_variant == CollisionVariant.DEAD else True)
                },
                "destinations": [_dest(_p(), ax, ay), _dest(_p(True), 0, 0) ]
                } for ax in [-1, 0, 1] for ay in [-1, 0, 1] ]
    
        car0_aut["edges"].append({
            "location": locstr % "send_trajectory_to_map",
            "action": "check_if_valid",
            # "guard": { "exp": Je.Not(Je.And(Je.Eq(car0dx, 0), Je.Eq(car0dy, 0))) },
            "destinations": [
                { "location": locstr % "wait_for_result", "assignments": [
                    { "ref": car0finished, "value": False },
                    { "ref": "map_x", "value": car0x },
                    { "ref": "map_y", "value": car0y },
                    { "ref": "map_dx", "value": car0dx },
                    { "ref": "map_dy", "value": car0dy },
                ] }
            ]
        })
    
        car0_aut["edges"].append({
            "location": locstr % "wait_for_result",
            "action": "is_valid",
            "destinations": [{
                "location": locstr % "sync_step",
                "assignments": [
                    { "ref": car0x, "value": "map_x" },
                    { "ref": car0y, "value": "map_y" },
                    { "ref": "map_x", "value": 0 },
                    { "ref": "map_y", "value": 0 },
                    { "ref": "map_dx", "value": 0 },
                    { "ref": "map_dy", "value": 0 },
                ]}
            ]
        })
    
        if self.collision_variant == CollisionVariant.DEAD:
            car0_aut["edges"].append({
                "location": locstr % "wait_for_result",
                "action": "is_invalid",
                "destinations": [{
                    "location": locstr % "sync_step",
                    "assignments": [
                        { "ref": car0x, "value": "map_x" },
                        { "ref": car0y, "value": "map_y" },
                        { "ref": car0dx, "value": 0 },
                        { "ref": car0dy, "value": 0 },
                        { "ref": "map_x", "value": 0 },
                        { "ref": "map_y", "value": 0 },
                        { "ref": "map_dx", "value": 0 },
                        { "ref": "map_dy", "value": 0 },
                    ] + ([ { "ref": car0crashed, "value": True } ])
                }]
            })
        elif self.collision_variant == CollisionVariant.BONET:
            car0_aut["edges"].append({
                "location": locstr % "wait_for_result",
                "action": "is_invalid",
                "destinations": [{
                    "location": locstr % "sync_step",
                    "assignments": [
                        { "ref": car0x, "value": "map_x" },
                        { "ref": car0y, "value": "map_y" },
                        { "ref": car0dx, "value": 0 },
                        { "ref": car0dy, "value": 0 },
                        { "ref": "map_x", "value": 0 },
                        { "ref": "map_y", "value": 0 },
                        { "ref": "map_dx", "value": 0 },
                        { "ref": "map_dy", "value": 0 },
                    ]}
                ]
            })
        else:
            car0_aut["edges"].append({
                "location": locstr % "wait_for_result",
                "action": "is_invalid",
                "destinations": [{
                    "location": locstr % "sync_step",
                    "probability": { "exp": 1 if len(self._grid.start) == 1 else Je.Div(1, len(self._grid.start)) },
                    "assignments": [
                        { "ref": car0x, "value": sx },
                        { "ref": car0y, "value": sy },
                        { "ref": car0dx, "value": 0 },
                        { "ref": car0dy, "value": 0 },
                        { "ref": "map_x", "value": 0 },
                        { "ref": "map_y", "value": 0 },
                        { "ref": "map_dx", "value": 0 },
                        { "ref": "map_dy", "value": 0 },
                    ]
                } for (sx, sy) in self._grid.start ]
            })
    
        car0_aut["edges"].append({
            "location": locstr % "wait_for_result",
            "action": "passed_finish_line",
            "destinations": [{
                "location": locstr % "sync_step",
                "assignments": [
                    # { "ref": car0dx, "value": 0 },
                    # { "ref": car0dy, "value": 0 },
                    { "ref": car0x, "value": "map_x" },
                    { "ref": car0y, "value": "map_y" },
                    { "ref": car0finished, "value": True },
                    { "ref": "map_x", "value": 0 },
                    { "ref": "map_y", "value": 0 },
                    { "ref": "map_dx", "value": 0 },
                    { "ref": "map_dy", "value": 0 },
                ]}
            ]
        })

        car0_aut["edges"].append({
            "location": locstr % "wait_for_result",
            "action": "exceeded_speed_limit",
            "destinations": [{
                "location": locstr % "deadlocked",
                "assignments": [
                    { "ref": car0x, "value": "map_x" },
                    { "ref": car0y, "value": "map_y" },
                    # { "ref": car0dx, "value": 0 },
                    # { "ref": car0dy, "value": 0 },
                    { "ref": "map_x", "value": 0 },
                    { "ref": "map_y", "value": 0 },
                    { "ref": "map_dx", "value": 0 },
                    { "ref": "map_dy", "value": 0 },
                ]}
            ]
        })

        car0_aut["edges"].append({
            "location": locstr % "sync_step",
            "action": "increment_step",
            "destinations": [
                { "location": locstr % "sync_step", }, 
            ]
        })

        car0_aut["edges"].append({
            "location": locstr % "sync_step",
            "action": "synchronize",
            "destinations": [
                { "location": locstr % "wait_for_control" },
            ]
        })

        return car0_aut

    def _car_automata(self):
        return [ self._car_automaton(i) for i in range(self._num_cars) ]

    def _controller_automaton(self, car_id):
        enabled_accelerations = [(ax, ay) for ax in [-1, 0, 1] for ay in [-1, 0, 1]]
        #enabled_accelerations = [(0, 0)] #, (1, 1)]
        agent0_aut = {
            "name": "agent%d" % car_id,
            "locations": [ { "name": "agent%d_loc" % car_id } ],
            "initial-locations": [ "agent%d_loc" % car_id ],
            "edges":
                [ { "location": "agent%d_loc" % car_id, "action": self._accel_name(car_id, ax, ay), "destinations": [ { "location": "agent%d_loc" % car_id } ] } for (ax, ay) in enabled_accelerations ]
        }
        return agent0_aut

    def _controller_automata(self):
        return [ self._controller_automaton(i) for i in range(self._num_cars) ]

    def _step_counter_automaton(self):
        # all_finished = "car0_finished" if self._num_cars == 1 else Je.And(["car%d_finished" % i for i in range(self._num_cars)])

        counter_aut = {
            "name": "counter",
            "locations": [
                { "name": "counter0" },
                { "name": "counter1" },
                { "name": "counter2" },
            ],
            "initial-locations": [ "counter2" ],
            "edges": [
                {
                "location": "counter0",
                "action": "increment_step",
                "destinations": [ { "location": "counter1",
                    "assignments": [{ "ref": "step_indicator", "value": 1 } ] \
                            + ([] if self._horizon is None else [ { "ref": "horizon", "value": Je.Add("horizon", 1) } ])
                    } ]
                },
                {
                    "location": "counter1",
                    "action": "synchronize",
                    "destinations": [ { "location": "counter2", "assignments": [{"ref": "step_indicator", "value": 0}, {"ref": "step_completed", "value": True } ] } ]
                },
                {
                    "location": "counter2",
                    "action": "reset",
                    **({} if self._horizon is None else { "guard": { "exp": Je.Lt("horizon", self._horizon) } }),
                    "destinations": [ { "location": "counter0", "assignments": [ { "ref": "step_completed", "value": False } ] }, ]
                },
            ]
        }
        return counter_aut

    def generate(self, model_name, grid, speed_limits, horizon, controller):
        self._grid = grid
        self._speed_limits = speed_limits
        self._is_speed_limited = any([speed_limits[x][y] != 0 for x in range(len(speed_limits)) for y in range(len(speed_limits[x]))])
        self._horizon = horizon
        self._num_cars = 1

        bounds = self._bounds()

        automata = [ self._step_counter_automaton(), self._map_automaton_discrete()] + self._car_automata() + self._controller_automata()
        if controller != None:
            automata.append(controller)

        num_auts = len(automata)
        def car_idx(car_id):
            return 2 + car_id
        def controller_idx(car_id):
            return 2 + self._num_cars + car_id

        jani = {
            "jani-version": 1,
            "name": model_name,
            "type": "mdp",
            "features": [ "state-exit-rewards", "arrays" ],
            "actions": self._actions(),
            "constants": self._constants(),
            "variables": self._variables(bounds=bounds),
            "restrict-initial":  { "exp": Je.And(self._initial_state(bounds=bounds)) },
            "properties": self._properties(bounds=bounds),
            "automata": automata,
        }

        syncs = []

        assert(self._num_cars == 1)
        # TODO
        # if more than one car -> currently would need to enumerate all possible
        # action choices (assuming all cars controlled)
        # => should introduce additional action encoding the "start of a move"
        # (in car or controller automata); synchronize over this action
        for ax in [-1, 0, 1]:
            for ay in [-1, 0, 1]:
                apply_move_reset_counter = { "synchronise": [ None ] * num_auts, "result": self._accel_name(0, ax, ay) }
                apply_move_reset_counter["synchronise"][0] = "reset" # counter
                apply_move_reset_counter["synchronise"][2] = self._accel_name(0, ax, ay)
                apply_move_reset_counter["synchronise"][3] = self._accel_name(0, ax, ay)
                if (controller != None):
                    apply_move_reset_counter["synchronise"][4] = self._accel_name(0, ax, ay)
                syncs.append(apply_move_reset_counter)

        tmp = [ None ] * num_auts
        tmp[0] = "increment_step"
        for i in range(self._num_cars):
            tmp[car_idx(i)] = "increment_step"
        syncs.append( { "synchronise": tmp, "result": "increment_step" } )

        tmp = [ None ] * num_auts
        tmp[0] = "synchronize"
        for i in range(self._num_cars):
            tmp[car_idx(i)] = "synchronize"
        if controller != None:
            tmp[-1] = "synchronize"
        syncs.append( { "synchronise": tmp, "result": "synchronize" } )

        for i in range(self._num_cars):
            for a in [ "check_if_valid", "is_valid", "is_invalid", "passed_finish_line", "exceeded_speed_limit" ]:
                tmp = [ None ] * num_auts
                tmp[1] = a
                tmp[car_idx(i)] = a
                syncs.append( { "synchronise": tmp, "result": a } )

        if controller != None:
            for a in [ "check_if_valid", "is_valid", "is_invalid", "passed_finish_line", "exceeded_speed_limit" ]:
                tmp = [ None ] * num_auts
                tmp[1] = a
                tmp[-1] = a
                syncs.append( { "synchronise": tmp, "result": a } )

        jani["system"] = {
            "elements": [ { "automaton": aut["name"] } for aut in automata ],
            "syncs": syncs
        }

        return jani

if __name__ == "__main__":
    p = argparse.ArgumentParser(description="Jani racetrack instance generator")
    p.add_argument("racetrack", help="Racetrack instance file")
    p.add_argument("--out", "-o", default=None, help="Output jani file")
    p.add_argument("--horizon", type=int, default=None, help="If provided, a finite horizon model will be generated")
    p.add_argument("--prob", default="0.9", help="Success probability of accelerating")
    p.add_argument("--2d", dest="twod", action="store_true", default=False, help="Represent map as 2D array.")
    p.add_argument("--collision", choices=["dead", "bonet", "barto"], default="dead", help="Which wall-hit variant to generate.")
    p.add_argument("--noise", choices=["off", "nochange", "offbyone", "offbyonetrunc"], default="nochange", help="Which noise variant to generate.")
    p.add_argument("--parking", action="store_true", default=False, help="Requires the car to exactly hit the goal zone (i.e., not just passing through).")
    p.add_argument("--zero-velocity", action="store_true", default=False, help="Requires the car really park in goal zone (i.e., hitting goal zone + having  there).")
    p.add_argument('--storm-compatible', action="store_true", default=False, help="Generate Storm compatible model.")
    p.add_argument('--resolve-nondeterminism', '-d', action="store_true", default=False, help="Use car controller.")
    p.add_argument('--global-speed-limit', type=int, default=None, help="Enforces the same speed limit to all cells.")
    p.add_argument("--no-speed-limit", default=False, action="store_true", help="Disable speed limit (even if defined in map file).")
    p.add_argument("--initial-dx",  type=int, default=0, help="Initial velocity in x direction.")
    p.add_argument("--initial-dy", type=int, default=0, help="Initial velocity in y direction.")
    def start_point(s):
        choices=["astrack", "first", "random", "allnonwalls"]
        if s.lower() in choices:
            return s.lower()
        result = []
        positions = s.split(",")
        for strpos in positions:
            pos = re.match("^(\d+):(\d+)$", strpos.strip())
            if pos is None:
                raise argparse.ArgumentTypeError("start point must either be a comma separated list of board positions ``x:y'', or in %s" % choices)
            result.append((int(pos.group(1)), int(pos.group(2))))
        return result
    p.add_argument('--start-point', type=start_point, default="astrack", help="Build model assuming single start point.")
    p.add_argument("--individual-jani-for-each-start-point", "--individual", default=False, action="store_true", help="Generate an individual Jani file for each start point.")
    # p.add_argument("--discretization-factor", "-f", default=1, type=int, help="Discretization factor (the map will effectively be enlargened by the provided factor, acceleration will be scaled accordingly)")

    args = p.parse_args()
    args.twod = args.twod and not args.storm_compatible

    (grid, speed_limits) = parse_grid_file(args.racetrack, 1, 0 if args.no_speed_limit else args.global_speed_limit) #args.discretization_factor)
    if args.start_point == "astrack":
        pass
    elif args.start_point == "first":
        grid.start = grid.start[:1]
    elif args.start_point == "random":
        grid.start = [grid.start[random.randint(0, len(grid.start) - 1)]]
    elif args.start_point == "allnonwalls":
        grid.start = [(x, y) for x in range(grid.dimx) for y in range(grid.dimy) if grid.get(x, y) != CellType.WALL]
    else:
        for (x, y) in args.start_point:
            if x < 0 or x >= grid.dimx or y < 0 or y >= grid.dimy:
                sys.stderr.write("Error: Start point (%d, %d) lies outside the grid!\n" % (x, y))
                sys.exit(1)
            elif grid.get(x, y) == CellType.WALL:
                sys.stderr.write("Error: There is a wall at start point (%d, %d)!\n" % (x, y))
                sys.exit(1)
        grid.start = args.start_point

    car_controller = None
    if args.resolve_nondeterminism:
        if len(grid.start) > 1:
            sys.stderr.write("Car controller implementation currently assumes a single start point")
            sys.exit(1)
        from controller import CarControllerAutomaton
        car_controller = CarControllerAutomaton(grid, 0).jani()
    
    gen = RacetrackGenerator(
            not args.twod,
            CollisionVariant.from_string(args.collision),
            args.parking,
            args.zero_velocity,
            NoiseVariant.from_string(args.noise),
            args.initial_dx, 
            args.initial_dy, 
            fractions.Fraction(decimal.Decimal(args.prob)) if args.noise != "off" else F_ONE,
            storm_compatible=args.storm_compatible)

    args.out = args.out or os.path.basename(args.racetrack) + ".jani"
    if len(grid.start) > 1 and args.individual_jani_for_each_start_point:
        padding = "%%0%dd" % int(math.ceil(math.log10(max(grid.dimx, grid.dimy))))
        if not args.out.endswith(".jani"):
            args.out += ".jani"
        args.out = args.out[:-5] + "-x%sy%sacc_x" % (padding, padding,)
        args.out = args.out + str(gen.initial_dx).zfill(2)+"acc_y"+str(gen.initial_dy).zfill(2)+".jani"
        possible_start = grid.start
        for (x, y) in possible_start:
            grid.start = [(x, y)]
            jani = gen.generate("racetrack", grid, speed_limits, args.horizon, car_controller)
            with open(args.out % (x, y), "w", encoding='utf-8') as f:			#, encoding='utf-8'
                f.write(json.dumps(jani, ensure_ascii=False, sort_keys=True, indent=2))		#indent=2
    else:
        jani = gen.generate("racetrack", grid, speed_limits, args.horizon, car_controller)
        if args.out == "--":
            print(json.dumps(jani, ensure_ascii=False, sort_keys=True, indent=2))	#indent=2
        else:
            with open(args.out, "w", encoding='utf-8') as f:			#, encoding='utf-8'
                f.write(json.dumps(jani, ensure_ascii=False, sort_keys=True, indent=2))		#indent=2

